#! /usr/bin/env python

# cpf2tree.py --- tree building and optimization for already parsed TDBSF
#                 search expressions
# Copyright (c) 2001 Florent Rougon
#
# This file is part of TDBSF.
#
# TDBSF is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# TDBSF is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with TDBSF; see the file COPYING. If not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA. */

import faststack
import expr2cpf
import string
import re
from tdbsf_exceptions import TDBSFGenericException

# The only exception defined by this module
class TDBSFRegexpCompilationError(TDBSFGenericException):
    ExceptionPrettyIdentifier = "Regular expression compilation error"

def hsort(a, b):
    if a[1] < b[1]: return -1
    elif a[1] == b[1]: return 0
    else: return 1

def one_op_and_one_regexp(a, b, stack):
    a[2].append(b)                      # a[2] is modified "in place"
    a[2].sort(hsort)                    # a[2] is modified "in place"
    stack.push([a[0], a[1] + b[1] + 2, a[2]]) # + 2 is somewhat arbitrary!

def build_node_sort(a, b, stack, op):
    heaviness = a[1] + b[1] + 2 # + 2 is somewhat arbitrary
    if a[1] <= b[1]:
        stack.push([op, heaviness, [a, b]])
    else:
        stack.push([op, heaviness, [b, a]])

def compile_regexp(pattern, flags_string):
    # We must insert "re." at the beginning of each flag name so that it
    # refers (I do not want to do an ugly `from re import *') to the right
    # object of the `re' module.
    if flags_string: # If flag(s) are provided for this regexp
        flags_list = string.split(flags_string, '|')
        for i in range(len(flags_list)):
            flags_list[i] = "re." + flags_list[i]

    try:
        if flags_string:
            regexp_object = re.compile(pattern,
                                       eval(string.join(flags_list, '|')))
        else:
            regexp_object = re.compile(pattern)
    except re.error, message:
        raise TDBSFRegexpCompilationError(message)
    return regexp_object

def build_regexp_leave(cpf_elt): # "leave" of the "tree" of operations
    pattern = cpf_elt[1]
    # The "heaviness" of the regexp is set to the length of its pattern.
    return ['r', len(pattern), compile_regexp(pattern, cpf_elt[2])]

def cpf2tree(cpf):
    stack = faststack.FastStack()

    i = 0
    while i < len(cpf):
        elt = cpf[i]
        op = elt[0] # operator like "&" or "!", or regexp indicated by "r"
        if op == 'r':
            stack.push(build_regexp_leave(elt))
        elif op in "|&":
            a = stack.pop()
            b = stack.pop()
            if (a[0] == op) and (b[0] == 'r'):
                one_op_and_one_regexp(a, b, stack)
            elif (b[0] == op) and (a[0] == 'r'):
                one_op_and_one_regexp(b, a, stack)
            elif a[0] == b[0] == op:
                a[2].extend(b[2])       # a[2] is modified "in place"
                a[2].sort(hsort)        # a[2] is modified "in place"
                                        # + 2 is somewhat arbitrary here:
                stack.push([a[0], a[1] + b[1] + 2, a[2]])
            else: # We cannot group the operations and their arguments
                build_node_sort(a, b, stack, op)
        elif op == '~':
            b = stack.pop()             # is a "regexp leave"
            n = int(stack.pop())        # is an integer
            a = stack.pop()             # is a "regexp leave"

                # Each regexp in a "near expression" increases a lot the
                # "heaviness" of the expression; +1 are here so that regexps
                # with "heaviness" 1 don't let the "heaviness" of the "near
                # expression" unchanged.
            c = ['~', (a[1]+1)*(b[1]+1), [a, b], [n]]

            # Get all the "near" expression at once. This enables to optimize
            # it afterwards by an optional inversion.
            while (i + 4 <= len(cpf)) \
                  and (cpf[i+1][0] in string.digits) \
                  and (cpf[i+2][0] == 'r') \
                  and (cpf[i+3][0] == '~'):
                new_regexp = build_regexp_leave(cpf[i+2])
                # Each regexp in a "near expression" increases a lot the
                # "heaviness" of the expression; +1 are here so that regexps
                # with "heaviness" 1 don't let the "heaviness" of the "near
                # expression" unchanged.
                c[1] = c[1]*(new_regexp[1]+1)
                c[2].append(new_regexp)
                c[3].append(int(cpf[i+1]))
                i = i + 3

            # The resulting c "leave" of the tree has the form:
            # ["~", xyz, [ [], [],..., [], []],
            #            [   n1, n2,..., np-1]]
            # where the p "[]" on the first line are "regexp leaves" (see
            # `build_regexp_leave' if you want to known what they look like).
                
            # Optimization: the last regexp should be *less* heavy than the
            # first one in the case of a "r1 ~n1 r2 ~n2 ... ~np-1 rp"
            # expression.
            if c[2][0][1] < c[2][-1][1]:
                c[2].reverse() # invert the order of the p regexps
                c[3].reverse() # invert the order of n1, n2, ..., np-1

            stack.push(c)               # Put our final "near leave" on the
                                        # stack.
        elif op == '!':
            a = stack.pop()
            stack.push(['!', a[1] + 2, a]) # The + 2 is somewhat arbitrary
        else: # elt should be a string of digits (the n or p in `a ~n b ~p c')
            stack.push(elt)

        i = i + 1
    return stack.pop()
        
def main():
    try:
        tree = cpf2tree(expr2cpf.gen_cpf_from_file("test_expr"))
    except (expr2cpf.TDBSFSyntaxError, TDBSFRegexpCompilationError), \
           exc_instance:
        import sys
        sys.stderr.write(exc_instance.complete_message() + '\n')
        return
    
    print tree

if (__name__ == "__main__"): main()
