#! /usr/bin/env python

# expr2cpf.py --- parser module for a TDBSF search expression
# Copyright (c) 2001 Florent Rougon
#
# This file is part of TDBSF.
#
# TDBSF is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# TDBSF is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with TDBSF; see the file COPYING. If not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA. */

import sys
import string
import faststack
import cStringIO
from tdbsf_exceptions import TDBSFGenericException

global input, stack, elts, cpf, line

# The only exception defined by this module
class TDBSFSyntaxError(TDBSFGenericException):
    ExceptionPrettyIdentifier = "Syntax error in match expression"

line = 1

class FileInput:
    def __init__(self, name):
        self.file = open(name, 'r')
    def read(self, n):
        return self.file.read(n)
    def seek(self, offset, whence=0):
        return self.file.seek(offset, whence)
    def close(self):
        return self.file.close()

class StringInput:
    def __init__(self, string):
        self.StringIO = cStringIO.StringIO(string)
    def read(self, n):
        return self.StringIO.read(n)
    def seek(self, offset, whence=0):
        return self.StringIO.seek(offset, whence)
    def close(self):
        return self.StringIO.close()

class Output:
    def __init__(self): pass
    def write(self, string):
        sys.stdout.write(string)

class EltProps:
    def __init__(self, name, cpf, prio, opening, closing):
        self.name = name
        self.cpf = cpf
        self.prio = prio
        self.opening = opening
        self.closing = closing
    def __str__(self):
        return "<EltProps: name : '" + self.name\
              + "', cpf : '" + self.cpf\
              + "', prio : '" + `self.prio`\
              + "', opening : '" + `self.opening`\
              + "', closing : '" + `self.closing` + "'>"

class Elt:
    def __init__(self, name, attrib=None):
        self.name = name
        self.attrib = attrib or {}
    def __str__(self):
        return "<Elt: '" + self.name + "', " + str(self.attrib) + ">"

def err(string):
    global line
    raise TDBSFSyntaxError("%s (line %d)" % (string, line))

# ****************************************************************************
# *                               The grammar                                *
# ****************************************************************************

def expr():
    term()
    while reco("|"): term()

def term():
    factor()
    while reco("&"): factor()

def factor():
    if reco("("):
        expr()
        if not reco(")"):
            err("missing close-parenthesis")
    elif reco("!"):
        factor()
    else:
        atom()

def atom():
    regexp()
    while reco("~"):
        integer()
        atom()

# Does NOT tolerate whitespace at the beginning
def integer():
    global cpf
    str = []

    while 1:
        c = input.read(1)
        if not c: break
        elif c in string.digits:
            str.append(c)
        else:
            input.seek(-1,1)
            break
    if str:
        cpf.append(string.join(str, ""))
    else:
        err("`~' must be immediately followed by an integer")

def regexp():
    global cpf, line, input
    pattern_str = []
    flags_str = []
    
    # The beginning of the read string defines a pattern. Then, if this string
    # is a quoted one and contains a non-escaped `%', we'll get the flags the
    # user wants to apply with the pattern. Remember:
    #     re.compile(pattern[,flags])
    str = pattern_str
    
    while 1:
        c = input.read(1)
        if not c: err("regular expression expected at end of file")
        elif c in ' \t': pass           # ignore spaces and tabs before regexps
        elif c == '\n': line = line + 1 # as well as newlines
        else:
            break

    if c == '"': # Quoted regexp
        while 1:
            c = input.read(1)
            if not c: err("quoted regexp not terminated at end of file")
            elif c == '\\':             # escape character
                c = input.read(1)
                if not c:
                    err("lonely `\\' in quoted regexp at end of file")
                else:
                    str.append(c)
            elif c == '"':
                break
            elif c == '%': # Next characters belong to the "flags" part
                str = flags_str
            elif c == '\n':
                line = line + 1
                str.append(c)
            else: str.append(c)
    elif c in "()|&!~": # No regexp begins with such chars
        err("expected a regexp, but found `%s'" % c)
    else: # Unquoted regexps may contain raw '\' and '"' characters
        str.append(c) # We have already read the regexp's 1st character
        while 1:
            c = input.read(1)
            if not c: break
            elif c in " \t\n)|&~": # These chars mark the end of an unquoted
                                   # regexp
                input.seek(-1,1)
                break
            elif c in "(!": err("regexp terminated by `%s'" % c)
            else: str.append(c)

    # 'r' indicates a regexp
    cpf.append(('r', string.join(pattern_str, ""),
                string.join(flags_str, "")))

def handle(elt_name, attrib=None):
    global stack, elts

    # Opening elements are put on the top of the stack
    if elts[elt_name].opening:
        stack.push(Elt(elt_name, attrib))
    else:
        # Conditional pop
        while elts[elt_name].prio <= elts[stack[0].name].prio:
            pop_elt()
        # Closing element? (e.g., closing parenthesis)
        if elts[elt_name].closing:
            stack_top = elts[stack[0].name]
            # Yes. Is the matching opening element present?
            if stack_top.opening and (stack_top.opening == elt_name):
                pop_elt()
            else:
                err("got `%s' without matching `%s'"
                    %(elt_name, elts[elt_name].closing))
        else:
            stack.push(Elt(elt_name, attrib))

# Pop an element from the stack and update the CPF
def pop_elt():
    global stack, cpf, elts

    cpf_code = elts[stack.pop().name].cpf
    if cpf_code:
        cpf.append(cpf_code)
    
def reco(str):
    global cpf, line

    while 1:
        c = input.read(1)
        if not c: return 0
        elif c in " \t": pass           # ignore spaces and tabs before tokens
        elif c == '\n': line = line + 1 # as well as newlines
        else:
            input.seek(-1,1)
            break
    n = len(str)
    # WARNING: if `str' contains newlines, the caller must increase the global
    # variable `line' if `reco' returns 1.
    strfich = input.read(n)
    if str == strfich:
        handle(str)
        return 1
    else:
        input.seek(-len(strfich), 1)
        return 0

def init_elts():
    global elts
    elts = { 'BOF': EltProps("BOF", "", 0, "EOF", 0),
             '(': EltProps("(", "", 0, ")", 0),
             'EOF': EltProps("EOF", "", 1, 0, "BOF"),
             ')': EltProps(")", "", 1, 0, "("),
             '|': EltProps("|", "|", 2, 0, 0), # OR
             '&': EltProps("&", "&", 3, 0, 0), # AND
             '!': EltProps("!", "!", 4, 0, 0), # NOT
             '~': EltProps("~", "~", 5, 0, 0)  # NEAR
             }

def gen_cpf():
    global input, stack, cpf, line

    init_elts()
    stack = faststack.FastStack()
    stack.push(Elt("BOF")) # Beginning Of File
    cpf = []

    expr()

    while 1:
        c = input.read(1)
        if not c:
            handle("EOF")
            break
        elif c in " \t": pass
        elif c == '\n': line = line + 1 
        else: err("data found after a valid expression")

    input.close()
        
    return cpf

def gen_cpf_from_file(filename):
    global input

    input = FileInput(filename)
    return gen_cpf()

def gen_cpf_from_string(string):
    global input

    input = StringInput(string)
    return gen_cpf()

def main():
    global cpf

    try:
        gen_cpf_from_string(r'''a & (bsdqqsd |!(c ~32 h))    & defg\y |
"cba\"g\\\%e()~%I|L|X" &
  (ghi ~3 abc ~4 def ~5 "kjh()!") | "d\e\""''')
    except TDBSFSyntaxError, e:
        sys.stderr.write(e.complete_message() + '\n')
        return
    output = Output()
    output.write(str(cpf) + '\n')
    
if (__name__ == "__main__"): main()
