#! /usr/bin/env python

# tdbsf_api.py --- module performing the searches through TDBSF database files
# Copyright (c) 2001 Florent Rougon
#
# This file is part of TDBSF.
#
# TDBSF is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# TDBSF is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with TDBSF; see the file COPYING. If not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA. */

"""API for the Trivial DataBase Searching Facility.

This module provides a function (`lookup_expr') to find records that match
an expression among a set of files conforming to a TDB format.

"""

import string
import re
import glob
import os.path
import sys
import expr2cpf
import cpf2tree
from tdbsf_exceptions import TDBSFGenericException

# One handy global variable
number_of_matches = 0

# Exceptions defined by this module
class TDBSFInvalidDBFileFormatError(TDBSFGenericException):
    ExceptionPrettyIdentifier = "Invalid database file format"
    
class TDBSFInvalidWhereArg(TDBSFGenericException):
    ExceptionPrettyIdentifier = "Invalid `where' argument to `lookup_expr'"
    
class TDBSFInvalidNearArg(TDBSFGenericException):
     ExceptionPrettyIdentifier = 'Invalid "near" (`~\') argument'

# The following characters will be considered word-constituent for near
# expressions (e.g. `foo ~3 "bars+" ~5 quux') . I did not include the
# uppercase  because it is not in the ISO-8859-1 character set (it is part of
# ISO-8859-15).
word_chars = string.letters + "-ԙۚ"
    

def locate_records(filename, filecontents, boh_re_obj, eoh_re_obj):
    """Locate TDB-type records defined by `boh' and `eoh'XXXregexps objects
    in `filecontents'."""

    res = []
    index = 0
        
    while 1:
        boh_match_object = boh_re_obj.search(filecontents, index)
        if not boh_match_object:
            break

        eoh_match_object = eoh_re_obj.search(filecontents,
                                             boh_match_object.end())
        if not eoh_match_object:
            raise TDBSFInvalidDBFileFormatError(
                  "the record beginning at position %d in `%s' has \
no `end of header' delimiter" % (boh_match_object.start(), filename))

        res.append((boh_match_object.span(), eoh_match_object.span()))
        index = eoh_match_object.end()

    return res

def try_regexp_matches(string, lower_limit, upper_limit, regexps,
                       max_offsets):
    while lower_limit < upper_limit:
        match = regexps[0][2].search(string, lower_limit)
        if match:
            # Absolute positions in `string' of the match start and end
            match_span = match.span()
            # Did the match start at a "legal" position?
            if (match_span[0] < upper_limit):
            # If max_offsets is [], `regexps' should have only one element,
            # the last regexp of the near expression, which has just matched.
               if not max_offsets:
                   return [match_span]
               # Go further trough the "near chain" of regexps
               next_res = near_next_regexp(string, match_span[0],
                                           regexps[1:], max_offsets)
               if next_res:
                   next_res.append(match_span)
                   return next_res
            # Look for another match for this regexp beginning further than
            # this one.
            lower_limit = match_span[0] + 1
        else:
            break

    return 0                            # The near expression didn't match

def near_next_regexp(string, anchor, regexps, max_offsets):
    global word_chars

    # If anchor is in a word, put it at the beginning of that word.
    if string[anchor] in word_chars:
        while (anchor > 0) and (string[anchor-1] in word_chars):
            anchor = anchor - 1

    lower_limit = anchor
    for i in range(max_offsets[0]):
        while (lower_limit > 0) and (string[lower_limit-1] not in word_chars):
            lower_limit = lower_limit - 1 
        while (lower_limit > 0) and (string[lower_limit-1] in word_chars):
            lower_limit = lower_limit - 1
        if lower_limit == 0:            # Don't do useless loops
            break

    upper_limit = anchor
    while (upper_limit < len(string)) \
          and (string[upper_limit] in word_chars):
        upper_limit = upper_limit + 1
    for i in range(max_offsets[0]):
        while (upper_limit < len(string)) \
              and (string[upper_limit] not in word_chars):
            upper_limit = upper_limit + 1
        while (upper_limit < len(string)) \
              and (string[upper_limit] in word_chars):
            upper_limit = upper_limit + 1
        if upper_limit == len(string):
            break                       # Don't do useless loops

    return try_regexp_matches(string, lower_limit, upper_limit,
                              regexps, max_offsets[1:])

def test_expr(test_tree, string):
    """Test if there is match of `regexp' in `string'."""
    op = test_tree[0] # operator, or 'r' indicating a regexp
    if op == '|':
        for or_arg in test_tree[2]:
            res = test_expr(or_arg, string)
            if res:
                return res
        return 0
    elif op == '&':
        res = []
        for and_arg in test_tree[2]:
            and_arg_res = test_expr(and_arg, string)
            if not and_arg_res:
                return 0
            # ! operations return only 0 or 1; they don't return any match span
            # information.
            elif and_arg_res != 1:
                res.append(and_arg_res)
        return ['&', res]
    elif op == '!':
        return not test_expr(test_tree[2], string)
    elif op == '~':
        regexps = test_tree[2]
        max_offsets = test_tree[3]
        # Verify user input
        for i in max_offsets:
            if i < 1:
                raise TDBSFInvalidNearArg(
                    "each word count `n_i' in a near expression \
`a_1 ~n_1 a_2 ~n_2 ... ~n_p a_(p+1)' must be strictly positive")
        # Recurse to evaluate the near expression
        res = try_regexp_matches(string, 0, len(string),
                                 regexps, max_offsets)
        if res:
            return ['~', res]
        else:
            return 0
        
    else: # op must be 'r' (regexp)
        # Beware: this will return a MatchObject instance or None, *not* 1 or
        # 0.
        res = test_tree[2].search(string)
        if res:
            return res.span()
        else:
            return 0

def add_to_result(res, records, record_index, match_place, match_pos, offset):
    """Update `res' after a sucessful test expression.

    Update `res' indicating a match in `filename' for the record identified
    by `record_index' in `records'.
    XXX incomplete

    """

    global number_of_matches

    number_of_matches = number_of_matches + 1
    
    res.append((
        (
        # Beginning of the header of the matching record (just before
        # the "boh" part)
        records[record_index][0][0] + offset,
        # Beginning of the header of the matching record (just after
        # the "boh" part)
        records[record_index][0][1] + offset,
        # Enf of the header of the matching record (just before the
        # "eoh" part)
        records[record_index][1][0] + offset,
        # Beginning of the record's body (just after the header and
        # its "eoh" part)
        records[record_index][1][1] + offset,
        # End of the matching record (i.e., beginning of the next
        # record or end of file if the matching record is the file's
        # last record)
        records[record_index+1][0][0] + offset
        ),
        (
        # Matched regexps positions (primarily intended for
        # fontification purposes)
        match_place, match_pos
        )
        ))

def try_to_guess_eol_type(str):
    # eol = End Of Line
    max_bytes_check_count = 400
    max_eol_check_count = 3

    beg = str[:max_bytes_check_count]

    # Try to find the end-of-line style (CR, LF or CRLF) using the same
    # algorithm as in GNU Emacs 21.1 (file "coding.c").
    eol_type = "undecided"
    eols_found = 0
    cr_or_lf_re = re.compile("[\r\n]")
    index = 0
    
    while (eols_found < max_eol_check_count):
        substring = beg[index:]
        mo = cr_or_lf_re.search(substring)

        if (mo == None): break          # nothing found until the end of the
                                        # string
        ms = mo.start()
        me = mo.end()

        index = index + me
        eols_found = eols_found + 1
        if substring[ms:me] == '\n':
            this_eol_type = "LF"
        elif ((me == len(substring)) or substring[ms+1:ms+2] != '\n'):
            this_eol_type = "CR"
        else:
            this_eol_type = "CRLF"
            index = index + 1       # Skip the following '\n'

        if (eol_type == "undecided"):
            # This is the first end-of-line
            eol_type = this_eol_type
        elif (this_eol_type != eol_type):
            eol_type = "undecided"
            break
    return eol_type

def try_to_convert_eols_to_lfs(str):
    eol_type = try_to_guess_eol_type(str)

    if (eol_type == "CR"):
        return string.replace(str, '\r', '\n')
    elif (eol_type == "CRLF"):
        return string.replace(str, '\r\n', '\n')
    else:
        # Either LF or inconsistent end-of-lines or no eol found within the
        # first characters of the file
        return str

def lookup_expr(expr, pattern, boh, boh_flags, eoh, eoh_flags, where,
                header_body_separator='', db_files_mode='r',
                consider_eol_as_lf=1, record_positions_offset=0):
    """Lookup an expression among a set of files.

    Lookup `expr' among all files matching `pattern'. The syntax for `pattern'
    is the one used by the glob module and therefore is close to the Unix
    shells' syntax.

    `where' specifies the string where `expr' is evaluated. It can take one
    of the following values:
      - "header"           to evaluate `expr' in the header of each record
      - "body"             to evaluate `expr' in the body of each record
      - "header_or_body"   to evaluate `expr' sequentially in the header and
        in the body of each record (a record is retained if `expr' matched
        either in its header or in its body)
      - "header_plus_body" to evaluate `expr' in a string which is, for each
        record, the concatenation of its header, the value of
        `header_body_separator' (which should be a string) and its body.

    `boh' is the regular expression used to identify the beginning of a
    header. Its counterpart, `eoh', marks the end of a header.
    The `eoh' and `boh' parts of a record are excluded from the search for
    `expression'.
    XXX db_files_mode
    XXX record_positions_offset
    XXX tell what is returned
    
    """

    global number_of_matches

    # Regular expression objects for the beginning of header and end of header
    # identification.
    boh_re_obj = cpf2tree.compile_regexp(boh, boh_flags)
    eoh_re_obj = cpf2tree.compile_regexp(eoh, eoh_flags)
    
    res = []

    for filename in glob.glob(pattern):
        file = open(filename, db_files_mode)
        rawfilecontents = file.read()
        file.close()

        if consider_eol_as_lf:
            filecontents = try_to_convert_eols_to_lfs(rawfilecontents)
        else:
            filecontents = rawfilecontents

        records = locate_records(filename, filecontents,
                                 boh_re_obj, eoh_re_obj)
        # To make the next loop easier
        records.append(((len(filecontents), 0), (0,0)))

        test_tree = cpf2tree.cpf2tree(expr2cpf.gen_cpf_from_string(expr))

        # Results concerning the currently examined  database file
        file_res = []

        # The loop below is repeated _inside_ the tests' clauses for
        # efficiency reasons.
        if where == "header":
            # - 1 because we added an artificial element to `records'
            for i in range(len(records) - 1):
                match_pos = test_expr(test_tree,
                                      filecontents[records[i][0][1]
                                                   :records[i][1][0]])
                if match_pos:
                    add_to_result(file_res, records, i, "header", match_pos,
                                  record_positions_offset)
        elif where == "body":
            # - 1 because we added an artificial element to `records'
            for i in range(len(records) - 1):
                match_pos = test_expr(test_tree,
                                      filecontents[records[i][1][1]
                                                   :records[i+1][0][0]])
                if match_pos:
                    add_to_result(file_res, records, i, "body", match_pos,
                                  record_positions_offset)
        elif where == "header_or_body":
            # - 1 because we added an artificial element to `records'
            for i in range(len(records) - 1):
                match_pos = test_expr(test_tree,
                                      filecontents[records[i][0][1]
                                                   :records[i][1][0]])
                if match_pos:
                    match_place = "header"
                else:
                    match_pos = test_expr(test_tree,
                                          filecontents[records[i][1][1]
                                                       :records[i+1][0][0]])
                    match_place = "body"
                if match_pos:
                    add_to_result(file_res, records, i, match_place,
                                  match_pos, record_positions_offset)
        elif where == "header_plus_body":
            # - 1 because we added an artificial element to `records'
            for i in range(len(records) - 1):
                match_pos = test_expr(test_tree,
                                      filecontents[records[i][0][1]
                                                   :records[i][1][0]]
                                      + header_body_separator +
                                      filecontents[records[i][1][1]
                                                   :records[i+1][0][0]])
                if match_pos:
                    add_to_result(file_res, records, i, "header_plus_body",
                                  match_pos, record_positions_offset)
        else: raise TDBSFInvalidWhereArg, \
              '`where\' should be one of "header", "body" and \
"header_and_body"'

        if file_res:                    # if a record in this file matched
            res.append([os.path.normpath(filename), file_res])
    return (number_of_matches, res)

def make_string(length, init):
    """Equivalent of Emacs' `make-string' function.
    
    Return a newly created string of length `length', with each element being
    `init'.

    """
    l = []
    for i in range(length):
        l.append(init)
    return  string.join(l, "")

def flo_db_test():
    print lookup_expr(
        r"""
    "[sS]haza+m !" & (bla | ("abc

azeaze" & "plonk !%IGNORECASE")) & "pou[e]t%I" ~3 def ~4 abc ~5 plonk ~6
"pou+i+c !" | fin ~3 header ~2 file2
    """
        , "d:/Florent\\python\\tdbsf/test_db/*.db",
        make_string(78, '=') + '\n', "",
        make_string(78, '-') + '\n', "",
        "header_or_body", '')

def roger_db_test():
    print lookup_expr(
        r"""
        "alumine%I"
        """
#          r"""
#          "ALUMINE +.*.* +DUNOD" & "Al2\(SO4\)3%I"
#          """
        , "d:/Florent\\python\\tdbsf/db_papa/*.[aA][sS][cC]",
        "  " + make_string(74, '') + '\n', "",
        "\n  " + make_string(74, '') + '\n', "",
        "header_or_body", '')

def test():
    try:
        flo_db_test()
    except (expr2cpf.TDBSFSyntaxError, TDBSFInvalidDBFileFormatError,
            TDBSFInvalidWhereArg, TDBSFInvalidNearArg,
            cpf2tree.TDBSFRegexpCompilationError), exc_instance:
        sys.stderr.write(exc_instance.complete_message() + '\n')
    except IOError:
        sys.stderr.write("I/O error.\n")
    except:
        sys.stderr.write(
            "TDBSF unhandled exception! Please remember precisely how you \
triggered it and report it to Florent!\n")
        
if __name__ == "__main__": test()
