;;; tdbsf-summary-mode.el --- TDBSF Summary mode

;; Copyright (c) 2001 Florent Rougon

;; Maintainer: Florent Rougon <flo@via.ecp.fr>

;; This file is part of TDBSF.

;; TDBSF is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; TDBSF is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with TDBSF; see the file COPYING. If not, write to the Free
;; Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Code:

(defvar tdbsf-summary-buffer-name "*TDBSF Results Summary*")

(defvar tdbsf-summary-mode-map (make-sparse-keymap)
  "Keymap for TDBSF Summary mode.")

(define-key tdbsf-summary-mode-map "\r" 'tdbsf-display-record-at)

(defvar tdbsf-summary-mode-hook nil
  "*Hook called by `tdbsf-summary-mode'.")

(defface tdbsf-summary-file-name-face
  '((
     ((class color))
     (:foreground "blue"))
    (t (:inverse-video t))              ; not tested
    )
  "Used to display a file name in tdbsf-summary-mode.")

(defun tdbsf-summary-mode ()
  "Major mode used to display and examine the results of a TDBSF search.
\\{tdbsf-summary-mode-map}"
  (interactive)
  (kill-all-local-variables)
  (use-local-map tdbsf-summary-mode-map)
  (mapcar 'make-local-variable
          '(tdbsf-results
            tdbsf-records-index
            tdbsf-session-id))
  (setq major-mode              'tdbsf-summary-mode
	mode-name               "TDBSF Summary"
	)
  (run-hooks 'tdbsf-summary-mode-hook)
  )

(defun tdbsf-prepare-db-file-buffer-and-display-it
  (buf record-index record-info-array global-record-info same-window)
  ;; Create a marker in `buf' for each record and build an array storing the
  ;; marker and the _global_ index for each record; set
  ;; `tdbsf-matches-overlays' to nil so that references to overlays previously
  ;; used for matches in `buf' are definitely gone; tag as well the buffer
  ;; with a session id so that later tdbsf queries giving different results
  ;; than the last one don't try to reuse these informations.
  (tdbsf-create-this-file-s-records-info-and-markers
   record-info-array buf (- record-index (aref global-record-info 2)))
  ;; Fontify all matches in `buf'
  (tdbsf-fontify-matches record-info-array buf)
  (tdbsf-bring-record-under-the-eyes record-index buf same-window))

(defun tdbsf-display-record (record-index same-window)
  (require 'tdbsf-db-file-mode)
  (let* ((global-record-info (aref
                              (with-current-buffer tdbsf-summary-buffer-name
                                tdbsf-records-index) record-index))
         (file (car (aref global-record-info 0)))
         (record-info-array (cdr (aref global-record-info 0)))
         (buf (get-file-buffer file))
         )
    (if buf
        (if (with-current-buffer buf
              (and (eq major-mode 'tdbsf-db-file-mode)
                   ;; Since we are here, `buf' is in tdbsf-db-file-mode, thus
                   ;; `tdbsf-session-id' is local to `buf'. Therefore, the
                   ;; following `boundp' tests its binding in `buf', even if
                   ;; `tdbsf-display-record' was called from the
                   ;; `tdbsf-summary-buffer-name' buffer, where
                   ;; `tdbsf-session-id' also has a binding.
                   (boundp 'tdbsf-session-id)))
            (if (string-equal (with-current-buffer buf tdbsf-session-id)
                              (with-current-buffer tdbsf-summary-buffer-name
                                tdbsf-session-id))
                (tdbsf-bring-record-under-the-eyes record-index
                                                   buf same-window)
              ;; We don't want to see the matches fontified by previous
              ;; sessions!
              (with-current-buffer buf
                (mapcar 'delete-overlay tdbsf-matches-overlays))
              ;; Warning: at this point, the overlays are "deleted" (i.e.
              ;; detached from `buf' but they are still referenced in
              ;; `tdbsf-matches-overlays'. The following function call ends
              ;; this situation by setting `tdbsf-matches-overlays' to nil.
              (tdbsf-prepare-db-file-buffer-and-display-it
               buf record-index record-info-array global-record-info
               same-window))
          (message "There is already a buffer visiting %s. Kill it before."
                   (file-name-nondirectory file))
          )
      (with-current-buffer (setq buf (find-file-noselect file))
        (tdbsf-db-file-mode))
      (tdbsf-prepare-db-file-buffer-and-display-it
       buf record-index record-info-array global-record-info same-window)
      )))

(defun tdbsf-display-record-at (location)
  "Display the TDBSF record corresponding to the line at LOCATION.

LOCATION is interpreted as a character position in the buffer named
according to `tdbsf-summary-buffer-name'. When called interactively,
LOCATION is set to the value of point."
  (interactive "d")
  (let* ((line                           ; counted starting at 0
         (save-excursion (beginning-of-line)(count-lines 1 (point))))
         (match-count (car
                       (with-current-buffer tdbsf-summary-buffer-name
                         tdbsf-results))))
    (unless (>= line match-count)
      (tdbsf-display-record line nil))))

(provide 'tdbsf-summary-mode)
