;;; tdbsf.el --- calling the TDBSF from Emacs, internal routines

;; Copyright (c) 2001 Florent Rougon

;; Maintainer: Florent Rougon <flo@via.ecp.fr>

;; This file is part of TDBSF.

;; TDBSF is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; TDBSF is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with TDBSF; see the file COPYING. If not, write to the Free
;; Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Code:

(defvar tdbsf-invocation-program-name
  "python"
  "XXX")

(defvar tdbsf-invocation-args
  '("/home/flo/src/tdbsf/core/tdbsf_front_end.py")
  "XXX")

(defvar tdbsf-raw-output-buffer-name "*TDBSF Raw Output*")
(defvar tdbsf-error-buffer-name "*TDBSF Error*")
(defvar tdbsf-temporary-buffer-name "*TDBSF Temporary Data*")

(defvar tdbsf-end-of-emacs-oriented-output-identifier
  "\n\n*** End of TDBSF Emacs-oriented output ***\n"
  "String output by the Emacs interface for the TDBSF after all the usefull
output.

This string is to ensure we have all the output in the buffer indicated by
`tdbsf-raw-output-buffer-name', because I am not sure that it is always the
case when a process sentinel gets a \"finished\" event.")

(defvar tdbsf-error-identifier
  "TDBSF error:\n\n"
  "String output by the Emacs interface for the TDBSF it catches an exception.
This string is the first thing that is output. It is then followed by an
explanation of the error.")

(defun tdbsf-insert-record-info (record-elt source-buffer-or-name
                                            extra-info-max-width)
  ;; Note: `extra-info-max-width' does *not* take the separator string
  ;;       inserted before the extra info into account (i.e. here, " ").
  ;;       It is the maximum width *of the extra info*.
  (insert " "
          (with-current-buffer source-buffer-or-name
            (save-restriction
              ;; The record's header
              (narrow-to-region (aref (car record-elt) 1)
                                (aref (car record-elt) 2))
              ;; We do not want the extra info to go beyond `fill-column'
              (if (< (+ (point-min) remaining-width) (point-max))
                  (narrow-to-region (point-min)
                                    (+ (point-min) remaining-width)))
              (goto-char (point-min))
              (end-of-line)
              (buffer-substring-no-properties (point-min) (point))))
          "\n"))

(defun tdbsf-process-output ()
  "XXX"
  (with-current-buffer tdbsf-raw-output-buffer-name
    ;; Check if the tdbsf call was ended by an exception
    (if (string-equal (buffer-substring-no-properties
                         (point-min) (+ (point-min)
                                        (length tdbsf-error-identifier)))
                        tdbsf-error-identifier)
        (progn
          ;; Put the error message in an appropriate buffer
          (let ((start (point-min))
                (end (- (point-max)
                        (length
                         tdbsf-end-of-emacs-oriented-output-identifier))))
            (with-current-buffer
                (get-buffer-create tdbsf-error-buffer-name)
              (erase-buffer)
              (insert-buffer-substring tdbsf-raw-output-buffer-name start end)))
          (display-buffer tdbsf-error-buffer-name)
          ;; FIXME: using `error' here causes trouble because of the function
          ;; is called by a process sentinel... What should I do?
          (message "TDBSF error -- details are in the buffer %s"
                   tdbsf-error-buffer-name))
      (let ((tdbsf-res (read (buffer-string)))
            (global-record-index -1))
        (if (zerop (car tdbsf-res))
            (message "TDBSF: no match found")
          (require 'tdbsf-summary-mode)
          (with-current-buffer
              (get-buffer-create tdbsf-summary-buffer-name)
            (tdbsf-summary-mode)
            ;; Buffer-local bindings
            (setq buffer-read-only nil
                  tdbsf-results tdbsf-res
                  tdbsf-records-index (make-vector
                                       ;; Total number of records that matched
                                       (car tdbsf-res)
                                       nil)
                  tdbsf-session-id (tdbsf-make-unique-session-id))
            (erase-buffer)
            ;; Loop over the files that matched
            (mapcar
             #'(lambda (file-elt)
                 (let* ((filename (file-name-nondirectory (car file-elt)))
                        (remaining-width (- fill-column
                                            (length filename) 1))
                        overlay start-of-overlay
                        (record-index-in-this-file -1))
                   ;; XXX I would prefer a "hidden buffer"
                   (with-current-buffer
                       (get-buffer-create tdbsf-temporary-buffer-name)
                     ;; The whole file
                     (insert-file-contents (car file-elt) nil
                                           nil nil t))
                   ;; Loop over the records that matched
                   (mapcar
                    #'(lambda (record-elt)
                        ;; Insert the name of the file corresponding to the
                        ;; current line and remember where it begins
                        (setq start-of-overlay (point))
                        (insert filename)
                        ;; Create and customize an overlay to fontify the file
                        ;; name
                        (overlay-put
                         (setq overlay
                               (make-overlay start-of-overlay (point)))
                         'face 'tdbsf-summary-file-name-face)
                        (overlay-put overlay 'evaporate t)
                        ;; Insert some info about the record corresponding to
                        ;; the current line
                        (tdbsf-insert-record-info record-elt
                                                  tdbsf-temporary-buffer-name
                                                  remaining-width)
                        ;; These are only *references*, not copies, thus I
                        ;; think there is no resource-eating problem
                        (aset tdbsf-records-index
                              (setq global-record-index
                                    (1+ global-record-index))
                              (vector file-elt record-elt
                                      (setq record-index-in-this-file
                                            (1+ record-index-in-this-file))))
                        )
                    (cdr file-elt))))
             (cdr tdbsf-res))
            ;; This binding is local to the buffer `tdbsf-summary-buffer-name'
            (setq buffer-read-only t)
            (goto-char (point-min))
            )
          (display-buffer tdbsf-summary-buffer-name)
          )))))

(defun tdbsf-process-sentinel (process event)
  (let ((i -1) (max-iterations 10) input-arrived)
    (when (string-equal event "finished\n")
      (with-current-buffer tdbsf-raw-output-buffer-name
        (while (and (< (setq i (1+ i)) max-iterations)
                    (not input-arrived))
          (if (string-equal
               tdbsf-end-of-emacs-oriented-output-identifier
               (buffer-substring
                (- (point-max)
                   (length tdbsf-end-of-emacs-oriented-output-identifier))
                (point-max)))
              (setq input-arrived t)    ; to end the loop ASAP
            (sleep-for 1)               ; wait a little bit and look again...
            )))
      (if input-arrived
          (tdbsf-process-output)
        (error "Could not see end of TDBFS output"))
      )
    ))

(defun call-tdbsf (expr pattern boh boh-flags eoh eoh-flags search-string
                        header-body-separator db-files-mode
                        &optional crlf-to-lf-in-args dont-consider-eol-as-lf)

  ;; Create if necessary and clear the " *TDBSF raw output*" buffer
  (with-current-buffer
      (get-buffer-create tdbsf-raw-output-buffer-name)
    (erase-buffer))

  (let ((tdbsf-process
         (apply 'start-process "tdbsf"
                tdbsf-raw-output-buffer-name
                (string-make-unibyte tdbsf-invocation-program-name)
                (mapcar 'string-make-unibyte
                        (append tdbsf-invocation-args
                                (if crlf-to-lf-in-args
                                    '("--crlf-to-lf-in-args"))
                                (if dont-consider-eol-as-lf
                                    '("--dont-consider-eol-as-lf"))
                                `("--emacs-friendly-output"
                                  "--boh-flags" ,boh-flags
                                  "--eoh-flags" ,eoh-flags
                                  "--db-files-mode" ,db-files-mode
                                  "--search-string" ,search-string
                                  "--header-body-separator"
                                  ,header-body-separator
                                  ,expr ,pattern ,boh ,eoh))))))
    (set-process-sentinel tdbsf-process 'tdbsf-process-sentinel)))

;; The two following functions are borrowed (and very slightly modified) from
;; Gnus v5.8.8, an excellent mail-and-news-reader for Emacs released under the
;; GNU GPL. They are defined in the file message.el of Gnus sources.
;; See http://www.gnus.org/ for more information.

(defun tdbsf-number-base36 (num len)
  (if (if (< len 0)
	  (<= num 0)
	(= len 0))
      ""
    (concat (tdbsf-number-base36 (/ num 36) (1- len))
	    (char-to-string (aref "zyxwvutsrqponmlkjihgfedcba9876543210"
				  (% num 36))))))

(defvar tdbsf-unique-id-char nil)

(defun tdbsf-make-unique-session-id ()
  ;; Don't use microseconds from (current-time), they may be unsupported.
  ;; Instead we use this randomly inited counter.
  (setq tdbsf-unique-id-char
	(% (1+ (or tdbsf-unique-id-char (logand (random t) (1- (lsh 1 20)))))
	   ;; (current-time) returns 16-bit ints,
	   ;; and 2^16*25 just fits into 4 digits i base 36.
	   (* 25 25)))
  (let ((tm (current-time)))
    (concat
     (if (memq system-type '(ms-dos emx vax-vms))
	 (let ((user (downcase (user-login-name))))
	   (while (string-match "[^a-z0-9_]" user)
	     (aset user (match-beginning 0) ?_))
	   user)
       (tdbsf-number-base36 (user-uid) -1))
     (tdbsf-number-base36 (+ (car   tm)
                             (lsh (% tdbsf-unique-id-char 25) 16)) 4)
     (tdbsf-number-base36 (+ (nth 1 tm)
                             (lsh (/ tdbsf-unique-id-char 25) 16)) 4)
     (number-to-string (random t)))))

(provide 'tdbsf)
